from django_erd_generator.definitions.models import (
    ModelArray as BaseModelArray,
    ModelDefinition,
)
from .models import Customer, Order, Product, Region


class ModelArray(BaseModelArray):
    def get_models(*args, **kwargs) -> BaseModelArray:
        arr = ModelArray(**kwargs)
        for cls in [Customer, Product, Order, Region]:
            arr.append(ModelDefinition(cls, **kwargs))
        return arr



from django_erd_generator.contrib.dialects import Dialect
from django_erd_generator.definitions.models import ModelArray

from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):
    help = "For one or more apps generate the code to generate an ERD in the syntax of choice."

    def add_arguments(self, parser):
        parser.add_argument(
            "-a",
            "--apps",
            required=False,
            default=None,
            help='The name of the apps which should be included in the ERD generated, these should be seperated by a comma, for example "shopping,polls". If no value is specified, all apps will be included.',
        )
        parser.add_argument(
            "-d",
            "--dialect",
            required=False,
            default="mermaid",
            help="The dialect which should be used, it should be either 'mermaid', 'plantuml' or 'dbdiagram'.",
        )
        parser.add_argument(
            "-o",
            "--output",
            required=False,
            default=None,
            help="The path to where the output of this content should be written, if no value is specified the output will be printed.",
        )

    def _parse_dialect(self, dialect: str) -> Dialect:
        valid = [i.value for i in Dialect]
        if dialect not in valid:
            valid_string = ", ".join(valid)
            err = f"{dialect} is not a valid choice, must be {valid_string}"
            raise CommandError(err)
        return Dialect(dialect)

    def _parse_apps(self, apps: str) -> list[str] | None:
        if not apps:
            return None
        return [i.strip() for i in apps.split(",")]

    def _write_output(
        self,
        output_destination: str,
        output_content: ModelArray,
    ) -> None:
        with open(output_destination, "w") as dst:
            dst.write(output_content.to_string())

    def handle(self, *args, **options):
        dialect = self._parse_dialect(options["dialect"])
        apps = self._parse_apps(options["apps"])
        output = options["output"]
        erd = ModelArray.get_models(apps, dialect=dialect)
        if output:
            self._write_output(output, erd)
        else:
            print(erd.to_string())



from django.core.management.base import BaseCommand
from django_erd_generator.utils.data_dictionary import DataDictionary


class Command(BaseCommand):
    help = "Generate a markdown data dictionary containing details of the models in one or more apps."

    def add_arguments(self, parser):
        parser.add_argument(
            "-a",
            "--apps",
            required=False,
            default=None,
            help='The name of the apps which should be included in the data dictionary generated, these should be seperated by a comma, for example "shopping,polls". If no value is specified, all apps will be included.',
        )
        parser.add_argument(
            "-o",
            "--output",
            required=False,
            default=None,
            help="The path to where the output of this content should be written, if no value is specified the output will be printed.",
        )

    def _parse_apps(self, apps: str) -> list[str] | None:
        if not apps:
            return None
        return [i.strip() for i in apps.split(",")]

    def handle(self, *args, **options):
        apps = self._parse_apps(options["apps"])
        output = options["output"]

        if output:
            DataDictionary.save_data_dictionary(apps=apps, path=output)
        else:
            print(DataDictionary.get_data_dictionary(apps=apps))



import re
from django.db import models, connection

from django_erd_generator.contrib.dialects import (
    FIELD_PATTERN_LOOKUP,
    PK_PATTERN_LOOKUP,
    Dialect,
)
from django_erd_generator.definitions.base import BaseArray, BaseDefinition
from django_erd_generator.definitions.relationships import Relationship


class FieldDefinition(BaseDefinition):
    def __init__(self, field: models.Field, dialect: Dialect = Dialect.MERMAID) -> None:
        self.django_field = field
        self.dialect = dialect
        self.col_name = self.django_field
        self.data_type = self.django_field
        self.primary_key = self.django_field

    @classmethod
    def get_relationship(
        self,
        field: models.Field,
        dialect: Dialect = Dialect.MERMAID,
    ) -> Relationship:
        rel_codes = ["one_to_many", "one_to_one", "many_to_one", "many_to_many"]
        if hasattr(field, "is_relation"):
            for rel_code in rel_codes:
                if getattr(field, rel_code, None):
                    return Relationship(field, rel_code, dialect=dialect)
        return None

    @classmethod
    def get_data_type(self, field: models.Field, dialect: Dialect) -> str:
        pattern = r"(\w+)\(([^)]+)\)"
        data_type = field.cast_db_type(connection)
        if data_type:
            matches = re.findall(pattern, data_type)
            args = None
            if matches:
                data_type, args = matches[0]
            if dialect is Dialect.MERMAID:
                # NOTE: MermaidJS erDiagram does not currently support spaces in either the field name,
                # or the data type. It incorrectly attempts to parse it as a comment.
                # More information: https://github.com/mermaid-js/mermaid/issues/1546
                data_type = data_type.replace(" ", "_")
            return {
                "data_type": data_type.lower(),
                "args": args,
            }
        return None

    @property
    def col_name(self) -> str:
        return self._col_name

    @col_name.setter
    def col_name(self, field: models.Field) -> None:
        self._col_name = field.attname

    @property
    def data_type(self) -> str:
        return self._data_type

    @data_type.setter
    def data_type(self, field: models.Field) -> None:
        self._data_type = self.get_data_type(field, self.dialect)

    @property
    def primary_key(self) -> bool:
        return self._primary_key

    @primary_key.setter
    def primary_key(self, field: models.Field) -> None:
        self._primary_key = field.primary_key

    def to_string(self) -> str:
        pattern = FIELD_PATTERN_LOOKUP[self.dialect]
        pk = PK_PATTERN_LOOKUP[self.dialect] if self.primary_key else ""
        col_name = self.col_name
        if self.dialect is Dialect.PLANTUML and self.primary_key:
            col_name = "*" + self.col_name
        return pattern.format(
            col_name=col_name,
            data_type=self.data_type["data_type"],
            primary_key=pk,
        )

    def __repr__(self) -> str:
        return self.to_string()


class FieldArray(BaseArray):
    pass



from django.db import models

from django_erd_generator.contrib.dialects import (
    REL_CODE_LOOKUP,
    REL_PATTERN_LOOKUP,
    Dialect,
)
from django_erd_generator.definitions.base import BaseArray, BaseDefinition


class Relationship(BaseDefinition):
    def __init__(
        self,
        field: models.Field = None,
        rel_code: str = None,
        dialect: Dialect = Dialect.MERMAID,
        from_model: str = None,
        from_field: str = None,
        to_model: str = None,
        to_field: str = None,
    ):
        self.dialect = dialect
        self.rel = rel_code
        if field:
            self.from_model = field.related_model.__name__
            self.from_field = field.related_model._meta.pk.attname
            self.to_model = field.model.__name__
            self.to_field = field.attname if hasattr(field, "attname") else field.name
            if self.rel in ["many_to_many", "one_to_one"]:
                self.to_field = field.model._meta.pk.attname
                self.from_field = field.related_model._meta.pk.attname
        else:
            self.to_model = to_model
            self.to_field = to_field
            self.from_model = from_model
            self.from_field = from_field

    def inverse(self) -> "Relationship":
        split_rel = self.rel.split("_")
        inverse_rel = "_".join([split_rel[-1], split_rel[1], split_rel[0]])
        return Relationship(
            from_field=self.to_field,
            from_model=self.to_model,
            to_field=self.from_field,
            to_model=self.from_model,
            rel_code=inverse_rel,
            dialect=self.dialect,
        )

    def to_string(self) -> str:
        rel_code = REL_CODE_LOOKUP[self.dialect][self.rel]
        pattern = REL_PATTERN_LOOKUP[self.dialect]
        return pattern.format(
            rel_code=rel_code,
            from_model=self.from_model,
            from_field=self.from_field,
            to_model=self.to_model,
            to_field=self.to_field,
        )


class RelationshipArray(BaseArray):
    pass



from typing import Any, Dict, List


class Table:
    def __init__(self, data: List[Dict[str, Any]] = []):
        self.data = data
        if len(self.data) == 0 or not self.data:
            return
        self.headers = list(data[0].keys())

    @property
    def headers(self) -> str:
        return "\n".join(
            [
                f"| {' | '.join(self._headers)} |",
                f"| {' | '.join(['---' for _ in self._headers])} | ",
            ]
        )

    @headers.setter
    def headers(self, headers: List[str]) -> None:
        self._headers = headers

    @property
    def body(self) -> str:
        return "\n".join(
            f"| {' | '.join([str(x) for x in i.values()])} |" for i in self.data
        )

    def __str__(self) -> str:
        return "\n".join([self.headers, self.body])



from django_erd_generator.contrib.dialects import Dialect


class BaseArray(list):
    def __init__(self, dialect: Dialect = Dialect.MERMAID):
        super().__init__()
        self.dialect = dialect

    def __repr__(self):
        return self.to_string()

    def to_string(self) -> str:
        return "\n".join(i.to_string() for i in self)


class BaseDefinition:
    def __repr__(self):
        return self.to_string()



from typing import Union
from django_erd_generator.contrib.dialects import (
    MODEL_PATTERN_LOOKUP,
    OUTPUT_PATTERN_LOOKUP,
    Dialect,
)
from django_erd_generator.definitions.base import BaseArray, BaseDefinition
from django_erd_generator.definitions.fields import FieldArray, FieldDefinition
from django_erd_generator.definitions.relationships import RelationshipArray
from django.db import models
import django.apps as d


class ModelDefinition(BaseDefinition):
    def __init__(self, model: models.Model, dialect: Dialect = Dialect.MERMAID) -> None:
        self.django_model = model
        self.dialect = dialect
        self.fields = self.django_model
        self.relationships = self.django_model
        self.name = model.__name__

    @property
    def fields(self) -> list[FieldDefinition]:
        return self._fields

    @fields.setter
    def fields(self, model: models.Model) -> None:
        valid_fields = FieldArray(dialect=self.dialect)
        for field in model._meta.get_fields():
            if field.concrete:
                definition = FieldDefinition(field, dialect=self.dialect)
                if definition.data_type:
                    valid_fields.append(definition)
        valid_fields.sort(key=lambda x: x.primary_key, reverse=True)
        self._fields = valid_fields

    @property
    def relationships(self) -> RelationshipArray:
        return self._relationships

    @relationships.setter
    def relationships(self, django_model: models.Model) -> None:
        valid_relationships = RelationshipArray(dialect=self.dialect)
        for field in django_model._meta.get_fields():
            relationship = FieldDefinition.get_relationship(field, dialect=self.dialect)
            if relationship and relationship.rel != "one_to_many":
                # NOTE: one_to_many and many_to_one are duplicated, so we only take one
                # of these values.
                valid_relationships.append(relationship)
        self._relationships = valid_relationships

    def to_string(self) -> str:
        return MODEL_PATTERN_LOOKUP[self.dialect].format(
            model_name=self.django_model.__name__,
            model_fields=self.fields.to_string(),
        )


class ModelArray(BaseArray):
    @classmethod
    def get_models(
        cls,
        valid_apps: Union[list[str], None] = None,
        dialect: Dialect = Dialect.MERMAID,
    ) -> "ModelArray":
        _dialect = Dialect(dialect)
        valid = cls(dialect=_dialect)

        models = [i for i in d.apps.get_models()]
        for model in models:
            if not valid_apps or (model._meta.app_label in valid_apps):
                valid.append(ModelDefinition(model, dialect=_dialect))
        return valid

    @property
    def relationships(self) -> RelationshipArray:
        models = [*self]
        valid = []
        unique = RelationshipArray(dialect=self.dialect)
        models.sort(key=lambda x: (len(x.relationships), x.name))
        for model in models:
            for relationship in model.relationships:
                if relationship.to_string() not in valid:
                    if relationship.inverse().to_string() not in valid:
                        valid.append(relationship.to_string())
                        unique.append(relationship)
        return unique

    def to_string(self) -> str:
        models_string = "\n".join([i.to_string() for i in self])
        return OUTPUT_PATTERN_LOOKUP[self.dialect].format(
            models=models_string,
            relationships=self.relationships.to_string(),
        )



import os
from unittest import TestCase

from django_erd_generator.contrib.dialects import Dialect
from .utils import ModelArray


class ModelDefinitionTestCase(TestCase):
    def setUp(self):
        self.models = ModelArray.get_models("tests")

    def test_field_names(self):
        expected_fields = [
            ["id", "first_name", "last_name"],
            ["id", "sku", "product_name", "product_code", "quantity", "price"],
            ["id", "customer_id", "product_id", "quantity", "order_total"],
            ["id", "name", "label"],
        ]
        for i, model in enumerate(self.models):
            field_names = [x.col_name for x in model.fields]
            names_exist = all([x in field_names for x in expected_fields[i]])
            self.assertEqual(names_exist, True)

    def test_field_types(self):
        expected_types = [
            ["integer", "text", "text"],
            ["integer", "text", "text", "integer", "decimal"],
            ["integer", "integer", "integer", "integer", "decimal"],
            ["integer", "text", "text"],
        ]
        for i, model in enumerate(self.models):
            field_types = [x.data_type["data_type"] for x in model.fields]
            types_correct = all([x in field_types for x in expected_types[i]])
            self.assertEqual(types_correct, True)

    def test_model_relationships(self):
        expected_relationships = {
            1: [
                ("Product", "Region", "many_to_many"),
            ],
            2: [
                ("Order", "Customer", "many_to_one"),
                ("Order", "Product", "many_to_one"),
            ],
        }
        for i, relationships in expected_relationships.items():
            target = self.models[i].relationships
            for j, (to_model, from_model, rel) in enumerate(relationships):
                self.assertEqual(target[j].from_model, from_model)
                self.assertEqual(target[j].to_model, to_model)
                self.assertEqual(target[j].rel, rel)


class ModelArrayTestCase(TestCase):
    def setUp(self):
        self._ = ModelArray.get_models("test")
        self.maxDiff = None

    def test_array_relationships(self):
        relationships = self._.relationships
        expected_relationships = [
            ("Product", "Region", "many_to_many"),
            ("Order", "Customer", "many_to_one"),
            ("Order", "Product", "many_to_one"),
        ]

        for i, (to_model, from_model, rel) in enumerate(expected_relationships):
            self.assertEqual(relationships[i].from_model, from_model)
            self.assertEqual(relationships[i].to_model, to_model)
            self.assertEqual(relationships[i].rel, rel)

    def test_model_array_dialect_render(self):
        test_dir = os.path.dirname(__file__)
        test_data = os.path.join(test_dir, "test_data")
        expected_render = {
            Dialect.MERMAID: open(os.path.join(test_data, "mermaid.txt")).read(),
            Dialect.PLANTUML: open(os.path.join(test_data, "plantuml.txt")).read(),
            Dialect.DBDIAGRAM: open(os.path.join(test_data, "dbdiagram.txt")).read(),
        }
        for dialect, expected in expected_render.items():
            model_arr = ModelArray.get_models("test", dialect=dialect)
            self.assertEqual(model_arr.to_string(), expected)



from enum import Enum


class Dialect(Enum):
    MERMAID = "mermaid"
    PLANTUML = "plantuml"
    DBDIAGRAM = "dbdiagram"
    MERMAID_FLOW = "mermaid_flow"


REL_CODE_LOOKUP = {
    Dialect.MERMAID: {
        "one_to_many": "||--|{",
        "one_to_one": "||--||",
        "many_to_one": "}|--||",
        "many_to_many": "}|--|{",
    },
    Dialect.PLANTUML: {
        "one_to_many": "||--|{",
        "one_to_one": "||--||",
        "many_to_one": "}|--||",
        "many_to_many": "}|--|{",
    },
    Dialect.DBDIAGRAM: {
        "one_to_many": "<",
        "one_to_one": "-",
        "many_to_one": ">",
        "many_to_many": "<>",
    },
    Dialect.MERMAID_FLOW: {
        "one_to_many": "o--x",
        "one_to_one": "o--o",
        "many_to_one": "x--o",
        "many_to_many": "x--x",
    },
}

REL_PATTERN_LOOKUP = {
    Dialect.MERMAID: '{to_model} {rel_code} {from_model}: ""',
    Dialect.MERMAID_FLOW: "{to_model} {rel_code} {from_model}",
    Dialect.DBDIAGRAM: "Ref: {to_model}.{to_field} {rel_code} {from_model}.{from_field}",
    Dialect.PLANTUML: "{to_model} {rel_code} {from_model}",
}

FIELD_PATTERN_LOOKUP = {
    Dialect.MERMAID: "  {data_type} {col_name} {primary_key}",
    Dialect.MERMAID_FLOW: "",
    Dialect.DBDIAGRAM: '  {col_name} "{data_type}" {primary_key}',
    Dialect.PLANTUML: "  {col_name}: {data_type}",
}

PK_PATTERN_LOOKUP = {
    Dialect.MERMAID: "pk",
    Dialect.MERMAID_FLOW: None,
    Dialect.DBDIAGRAM: "[primary key]",
    Dialect.PLANTUML: None,
}

MODEL_PATTERN_LOOKUP = {
    Dialect.MERMAID: "{model_name} {{\n{model_fields}\n}}",
    Dialect.MERMAID_FLOW: "{model_name}",
    Dialect.DBDIAGRAM: "Table {model_name} {{\n{model_fields}\n}}",
    Dialect.PLANTUML: "entity {model_name} {{\n{model_fields}\n}}",
}

OUTPUT_PATTERN_LOOKUP = {
    Dialect.MERMAID: "erDiagram\n{models}\n{relationships}",
    Dialect.MERMAID_FLOW: "flowchart\n{models}\n{relationships}",
    Dialect.DBDIAGRAM: "{models}\n{relationships}",
    Dialect.PLANTUML: "@startuml\n\n{models}\n{relationships}\n\n@enduml",
}



from unittest import TestCase

from django_erd_generator.contrib.dialects import Dialect
from django_erd_generator.definitions.fields import Relationship
from .models import Order
from .utils import ModelArray


class RelationshipTestCase(TestCase):
    def test_init_relationship_definition(self):
        field = Order._meta.get_field("customer")
        relationship = Relationship(field, "one_to_many")
        self.assertEqual(relationship.to_model, "Order")
        self.assertEqual(relationship.from_model, "Customer")
        self.assertEqual(relationship.rel, "one_to_many")

    def test_relationship_definition_dialect_render(self):
        field = Order._meta.get_field("customer")

        dialects = {
            Dialect.MERMAID: 'Order ||--|{ Customer: ""',
            Dialect.PLANTUML: "Order ||--|{ Customer",
            Dialect.DBDIAGRAM: "Ref: Order.customer_id < Customer.id",
        }

        for dialect, expected in dialects.items():
            rel_definition = Relationship(field, "one_to_many", dialect=dialect)
            self.assertEqual(expected, rel_definition.to_string().strip())


class RelationshipArrayTestCase(TestCase):
    def test_relationship_array_dialect_render(self):
        dialects = {
            Dialect.MERMAID: 'Order }|--|| Customer: ""\nOrder }|--|| Product: ""',
            Dialect.PLANTUML: "Order }|--|| Customer\nOrder }|--|| Product",
            Dialect.DBDIAGRAM: "Ref: Order.customer_id > Customer.id\nRef: Order.product_id > Product.id",
        }

        for dialect, expected in dialects.items():
            model = ModelArray.get_models("tests", dialect=dialect)[2]
            relationships = model.relationships.to_string()
            self.assertEqual(relationships, expected)



import subprocess


def get_git_commit() -> str:
    result = subprocess.run(
        ["git", "rev-parse", "HEAD"],
        capture_output=True,
        text=True,
        check=True,
    )
    return result.stdout.strip()



import django
from django.conf import settings

if not settings.configured:
    settings.configure(
        DEBUG_PROPAGATE_EXCEPTIONS=True,
        DATABASES={
            "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"},
            "secondary": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"},
        },
        SITE_ID=1,
        SECRET_KEY="not very secret in tests",
        USE_I18N=True,
        STATIC_URL="/static/",
        ROOT_URLCONF="tests.urls",
        TEMPLATES=[
            {
                "BACKEND": "django.template.backends.django.DjangoTemplates",
                "APP_DIRS": True,
                "OPTIONS": {
                    "debug": True,
                },
            },
        ],
        MIDDLEWARE=(
            "django.middleware.common.CommonMiddleware",
            "django.contrib.sessions.middleware.SessionMiddleware",
            "django.contrib.auth.middleware.AuthenticationMiddleware"
            "django.contrib.messages.middleware.MessageMiddleware",
        ),
        INSTALLED_APPS=(
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.sites",
            "django.contrib.staticfiles",
            "django_erd_generator",
            "tests.apps.TestsConfig",
        ),
        PASSWORD_HASHERS=("django.contrib.auth.hashers.MD5PasswordHasher",),
    )

django.setup()



from unittest import TestCase

from django_erd_generator.contrib.dialects import Dialect
from django_erd_generator.definitions.fields import FieldDefinition
from .models import Customer
from .utils import ModelArray


class FieldTestCase(TestCase):
    def test_init_field_definition(self):
        field = Customer._meta.get_field("id")
        field_definition = FieldDefinition(field)
        self.assertEqual(field_definition.col_name, "id")
        self.assertEqual(field_definition.data_type["data_type"], "integer")
        self.assertEqual(field_definition.primary_key, True)

    def test_field_definition_dialect_render(self):
        field = Customer._meta.get_field("id")

        dialects = {
            Dialect.MERMAID: "integer id pk",
            Dialect.PLANTUML: "*id: integer",
            Dialect.DBDIAGRAM: 'id "integer" [primary key]',
        }

        for dialect, expected in dialects.items():
            field_definition = FieldDefinition(field, dialect=dialect)
            self.assertEqual(expected, field_definition.to_string().strip())


class FieldArrayTestCase(TestCase):
    def test_field_array_dialect_render(self):
        dialects = {
            Dialect.MERMAID: [
                "integer id pk",
                "text first_name",
                "text last_name",
            ],
            Dialect.PLANTUML: [
                "*id: integer",
                "first_name: text",
                "last_name: text",
            ],
            Dialect.DBDIAGRAM: [
                'id "integer" [primary key]',
                'first_name "text"',
                'last_name "text"',
            ],
        }

        for dialect, expected in dialects.items():
            model = ModelArray.get_models("tests", dialect=dialect)[0]
            fields = model.fields.to_string()
            field_arr = [i.strip() for i in fields.split("\n")]
            self.assertEqual(" ".join(field_arr), " ".join(expected))



from collections import defaultdict
import os
from typing import Dict, List, Optional
from django_erd_generator.contrib.dialects import Dialect
from django_erd_generator.contrib.markdown import Table
from django_erd_generator.definitions.models import ModelArray, ModelDefinition
from django_erd_generator.utils.git import get_git_commit

MODEL_RENDER_TEMPLATE = """\
#### {model_name}

`{signature}`

{doc_string}

{table}
"""

APPS_RENDER_TEMPLATE = """\
### {app_name}

{models}
"""

DICTIOANRY_RENDER_TEMPLATE = """\
# {project_name} - Data Dictionary

Commit `{commit}`

---

## Table of Contents [#](#toc)

{toc}

---

## Modules [#](#modules)

{apps}

"""


class DataDictionary:
    model_array_class = ModelArray

    @classmethod
    def get_data_dictionary(
        cls, apps: Optional[List[str]] = None
    ) -> Dict[str, ModelDefinition]:
        model_map = {}
        for model in ModelArray.get_models(valid_apps=apps, dialect=Dialect.MERMAID):
            name = (
                f"{(cls := model.django_model)._meta.app_label}.models.{cls.__name__}"
            )
            model_map[name] = model
        return model_map

    @classmethod
    def get_apps_map(
        cls, apps: Optional[List[str]] = None
    ) -> Dict[str, List[ModelDefinition]]:
        data_dictionary = cls.get_data_dictionary(apps=apps)
        apps_map = defaultdict(list)
        for model_name, model in data_dictionary.items():
            app_label = model.django_model._meta.app_label
            setattr(model, "name", model_name)
            apps_map[app_label].append(model)
        return apps_map

    @classmethod
    def render_model(cls, model: ModelDefinition) -> str:
        doc_string = model.django_model.__doc__ or "No description provided."
        doc_string = doc_string.strip()

        fields = []
        for field in model._fields:
            meta = field.django_field.__dict__
            related = meta.get("related_model")
            fields.append(
                {
                    "pk": "✓" if meta.get("primary_key") else "",
                    "field_name": field._col_name,
                    "data_type": f"`{field._data_type['data_type'].replace('_', ' ') or ''}`",
                    "related_model": f"[{related.__name__}](#{related.__name__})"
                    if related
                    else related.__name__
                    if related
                    else "",
                    "description": meta.get("help_text", "").replace("\n", " "),
                    "nullable": "✓" if meta.get("null") else "",
                    "unique": "✓" if meta.get("unique") else "",
                    "choices": "✓" if meta.get("choices") else "",
                    "max_length": meta.get("max_length") or "",
                    "db_index": "✓" if meta.get("db_index") else "",
                }
            )
        table = Table(fields)

        model_name = model.name.split(".")[-1]
        signature = f"{model_name}({', '.join([i.name for i in model.django_model._meta.get_fields() if i.concrete])})"
        model_name = f"{model_name}[#](#{model.django_model.__name__})"

        if doc_string[: len(model_name)] == signature[: len(model_name)]:
            doc_string = ""

        return MODEL_RENDER_TEMPLATE.format(
            model_name=model_name,
            doc_string=doc_string,
            signature=signature,
            table=table,
        )

    @classmethod
    def generate_data_dictionary(cls, apps: Optional[List[str]] = None) -> str:
        apps_map = cls.get_apps_map(apps=apps)

        rendered_apps: List[str] = []

        project_name = "Django Project"
        settings_module = os.environ.get("DJANGO_SETTINGS_MODULE")
        if settings_module:
            project_name = settings_module.split(".")[0]

        toc = ["- [Table of Contents](#toc)", "- [Modules](#modules)"]
        for app, arr in apps_map.items():
            toc.append(f"  - [{app}](#{app})")

            apps_map[app] = sorted(arr, key=lambda x: x.django_model.__name__)

            rendered_models: List[str] = []
            for model in arr:
                model_name = model.django_model.__name__
                toc.append(f"    - [{model_name}](#{model_name})")
                rendered_models.append(cls.render_model(model))
            apps_map[app] = APPS_RENDER_TEMPLATE.format(
                app_name=app,
                models="\n".join(rendered_models),
            )
            rendered_apps.append(apps_map[app])

        return DICTIOANRY_RENDER_TEMPLATE.format(
            project_name=project_name,
            apps="\n".join(rendered_apps),
            commit=get_git_commit(),
            toc="\n".join(toc),
        ).replace("\n\n\n", "\n")

    @classmethod
    def save_data_dictionary(
        cls,
        path: str,
        apps: Optional[List[str]] = None,
    ) -> None:
        content = cls.generate_data_dictionary(apps=apps)
        with open(path, "w", encoding="utf-8") as f:
            f.write(content)
        print(f"Data dictionary saved to {path}")



from django.db import models


class TestModel(models.Model):
    id = models.AutoField(primary_key=True)

    class Meta:
        app_label = "tests"
        abstract = True


class Customer(TestModel):
    first_name = models.TextField()
    last_name = models.TextField()


class Product(TestModel):
    sku = models.TextField()
    product_name = models.TextField()
    product_code = models.TextField()
    quantity = models.IntegerField()
    price = models.DecimalField(max_digits=16, decimal_places=2)
    regions = models.ManyToManyField("Region")


class Order(TestModel):
    customer = models.ForeignKey(Customer, on_delete=models.CASCADE)
    product = models.ForeignKey(Product, on_delete=models.CASCADE)
    quantity = models.IntegerField()
    order_total = models.DecimalField(max_digits=16, decimal_places=2)


class Region(TestModel):
    name = models.TextField()
    label = models.TextField()
